package Freenet;
import Freenet.support.*;
import java.io.*;
/*
  This code is part of the Java Adaptive Network Client by Ian Clarke. 
  It is distributed under the GNU Public Licence (GPL) version 2.  See
  http://www.gnu.org/ for further details of the GPL.

  Explanation of Code Versions: 
    0.0.0      = Initial Description
    0.0.1      = API Specified
    0.x (x>0)  = Partial Implementation
    x.0 (x>0)  = Operational
		
  Requires Classes: java.io.* (JDK 1.1)
                    Callback (1.0)

 */

/**
 * This class creates a conduit between an InputStream and an 
 * OutputStream and when the feed method of this object is called
 * the InputStream will be fed into the output stream of this
 * object.
 *
 * @version 1.0
 * @author <A HREF="mailto:I.Clarke@strs.co.uk">Ian Clarke</A>
 **/

public class Conduit extends Thread
{
    private InputStream in;
    private OutputStream out;
    private Callback cb;
    private Callback icb;
    private ByteCounter counter;
    private long bytesToMove = -1;
    private long bytesMoved = 0;
    public long bytesMoved() {return bytesMoved;}

  /**
   * @param i The stream to read data from
   * @param o The stream to send data to
   **/
  public Conduit(InputStream i, OutputStream o)
    {
	this(i,o,null);
    }
  /**
   * @param i The stream to read data from
   * @param o The stream to send data to
   * @param counter The counter to count the bytes
   **/
    public Conduit(InputStream i, OutputStream o, ByteCounter counter) {
	this.in = i;
	this.out = o;
	this.counter = counter;
    }

   

  /**
   * Commences the transfer from Input to Output Streams asynchronously
   * @param cb An object implementing the Callback interface which will be 
   *           called when the stream finnishes.
   **/
  public void asyncFeed(Callback cb)
    {
	if (!isAlive()) {
	    this.cb=cb;
	    this.start();
	}
    }

    /**
     * Commences the transfer from Input to Output Stream asynchronously 
     * until a number of bytes have been transfered
     * @param cb An object implementing the Callback interface which will be 
     *           called when the stream finnishes.
     * @param icb An object of implementing the Callback interface which will
     *            called if the stream is interupted before reading all the 
     *            of bytes has finnished.
     * @param n  the number of bytes to read
     **/
    public void asyncFeed(Callback cb, Callback icb, long n)
    {
	if (!isAlive()) {
	    this.cb=cb;
	    this.icb = icb;
	    this.bytesToMove = n;
	    this.start();
	}
    }

  /**
   * Transfer from Input to Output Streams, doesn't return until transfer
   * complete
   * @depreciated Use asyncFeed() and wait() on one of the streams instead (this may even lock now)
   **/
  public void syncFeed()
    {
      if (!isAlive()) this.run();
    }

  public void run()
    {
      byte[] buffer = new byte[512];
      int bytes_read;
      Logger.log("Conduit.java","Conduit starting",Logger.DEBUG);
      synchronized(in) {
	  synchronized(out) {
	      try {
		  for(;;) {
		      bytes_read = in.read(buffer, 0, bytesToMove < 0 ||
					   (bytesToMove - bytesMoved >= buffer.length)
					   ? buffer.length
					   : (int) (bytesToMove - bytesMoved));
	      
		      if (bytes_read > 0)
			  out.write(buffer, 0, bytes_read);
		      bytesMoved = bytesMoved + bytes_read;


		      if (counter != null)
			  counter.count(bytes_read == -1 ? 0 : bytes_read);
	      
		      if (bytes_read == -1 || bytesMoved == bytesToMove) {
			  if (bytesToMove >= 0 && bytesMoved != bytesToMove && icb != null)
			      icb.callback();
			  else if (cb != null)
			      cb.callback();

			  if (counter != null)
			      counter.finished();
			  return;
		      }

		  }
	      }
	      catch (IOException e) {
		  Logger.log("Conduit.java",e.toString(),Logger.ERROR);
		  if (icb != null)
		      icb.callback();
		  else if (cb != null)
		      cb.callback();
	      } finally {
		  Logger.log("Conduit.java","Conduit finished",Logger.DEBUG);
		  // notify the streams
		  //	  out.notifyAll();

		  in.notifyAll();
	      }
	  }
      }
    }
}

