package Freenet;
import Freenet.support.*;
import java.net.*;
import java.io.*;
import java.io.EOFException;
/*
  This code is part of the Java Adaptive Network Client by Ian Clarke. 
  It is distributed under the GNU General Public Licence (GPL) 
  version 2.  See http://www.gnu.org/ for further details of the GPL.
 */

/**
 * Handles both sending and receiving messages on a connection.
 *
 * @author <A HREF="mailto:I.Clarke@strs.co.uk">Ian Clarke</A>
 * @author <a href="mailto:blanu@uts.cc.utexas.edu">Brandon Wiley</a>
 **/

public class ConnectionHandler extends Thread
{
  // Protected/Private Fields
    private static long ids=0;
    private long id;
  private Connection c;
  private MessageHandler mh;
  private volatile Long pongs = new Long(0);
    private boolean closed;
    private boolean waiting = true;

    protected class ConnectionCB implements Callback {
	public void callback() {
	    close();
	}
    }

  // Constructors
  public ConnectionHandler(Connection c, MessageHandler mh) {

      this.c = c;
      this.mh = mh;
      this.id = ++ids;
      Logger.log("ConnectionHandler.java#"+ id,"New connectionhandler with " + peer(),Logger.DEBUGGING);     
  }


    // Public Methods
  public void run() {
	
    message: do {
	RawMessage m=null;
	Message msg = null;
	try {
	    if(c==null) 
		break message;

	    while (!closed && c.in.available() == 0) {
	    	try {
	    	    Thread.sleep(50);
	    	} catch (InterruptedException e) {
	    	}
	    }
	    if (closed) break message;

	    m = new RawMessage(c.in);
	    closed = !m.isKeepAlive();

	    Logger.log("ConnectionHandler.java#"+ id,"Rawmessage:\n"+m,Logger.MINOR);
	    msg = MessageFactory.toMessage(m);

	} catch (InvalidMessageException e) {
	    Logger.log("ConnectionHandler.java#" + id,"Invalid message: " + e.toString(),Logger.MINOR);
	    continue message; 
	} catch (EOFException e) {
	    break message; // this stream is over
	} catch (IOException e) {
	    break message;
	}

	String logstr=m.messageType+" <- "+c.getPeerAddress();
	if(Logger.verbosity>0) logstr=Long.toHexString(msg.id)+" - "+logstr;
	Logger.log("ConnectionHandler.java#"+id,logstr,Logger.NORMAL);
	Logger.log("ConnectionHandler.java#"+id,"Message:\n"+msg,Logger.DEBUG);

	// set the receivedAt, source, and receivedWith fields
	Address source = null;
	try {
	    source = new Address(m.readField("Source"));
	} catch (Exception e) {}

	Logger.log("ConnectionHandler.java#"+id,"Got on " + c.getMyAddress() + " from " + (source != null ? c.getPeerAddress(source.listenPart()) : c.getPeerAddress()), Logger.DEBUGGING);

	msg.initSources(c.getMyAddress(),source == null ? null : c.getPeerAddress(source.listenPart()), m.isKeepAlive() ? this : null);

	// Handle
	mh.handle(msg);

	Logger.log("ConnectionHandler.java#"+id,"Finished with message",Logger.DEBUGGING);

	if (m.trailingFieldLength() > 0) {
	    Logger.log("ConnectionHandler.java#"+id,"Waiting on trailing field",Logger.DEBUGGING);
	    // wait() lets go of the input so the trailing field can be read by the handling code. The matching notify is (rather deviously) called from the Conduit class. 
	    synchronized(m.trailingFieldStream) {
		try {
		    m.trailingFieldStream.wait();
		} catch (InterruptedException e) {}
		Logger.log("ConnectionHandler.java#"+id,"I got notified!",Logger.DEBUGGING);
	    }
	}
    } while (!closed);

    Logger.log("ConnectionHandler.java#"+id,"Finished with connection - closing",Logger.DEBUGGING);
    synchronized(c.out) {
	Logger.log("ConnectionHandler.java#"+id,"Finished with connection - closing",Logger.DEBUGGING);
	c.close();
    }
    }

     /**
     * Sends a message using this connection
     * @param m            the message to send
     */   
    public void sendMessage(Message m) throws SendFailedException {
	sendMessage(m, null, null);
    }

    /**
     * Sends a message using this connection
     * @param m            the message to send
     * @param datatunnel   a stream to tunnel the data through (presumably so it can be copied to a file as well)
     * @param ByteCounter  a way to count how much data is sent
     */

    public void sendMessage(Message m, SplitOutputStream datatunnel, ByteCounter count) throws SendFailedException {
	Logger.log("ConnectionHandler.java#"+id,"Sending a message",Logger.DEBUGGING);
	RawMessage raw=m.toRawMessage();
        String logstr=raw.messageType+" -> "+c.getPeerAddress();
        if(Logger.verbosity>0) logstr=Long.toHexString(m.id)+" - "+logstr;
	Logger.log("ConnectionHandler.java#"+id,logstr,Logger.NORMAL);
	synchronized(c.out) {
	    try {
		raw.writeMessage(c.out);
	    } catch (IOException e) {
		throw new SendFailedException(c.getPeerAddress());
	    }
	    InputStream data = raw.getTrailing();
	    if (data != null) {
		ConnectionCB ccb;
		if (!raw.isKeepAlive())
		    ccb = new ConnectionCB();
		else
		    ccb = null;
		
		if (datatunnel!=null) {
		    datatunnel.addOutput(c.out,ccb);
		    (new Conduit(data, datatunnel, count)).asyncFeed(datatunnel,null,raw.trailingFieldLength()); // the null should be a callback that is called when it gets interupted before finnishing the data - at least now nothing happens, so the data will never get finalized (see Data.java and DataCallback)
		} else {
		    (new Conduit(data, c.out, count)).asyncFeed(ccb, null, raw.trailingFieldLength());
		}
	    } else if (!raw.isKeepAlive()) {
		this.close();
	    }
	} // the conduit is synchronized on the output stream too, so it should not start until now - there is a small danger here that another sendMessage would win the race instead of the conduit and start instead...
    }
    
    //I'm letting this force close now, maybe it should synchronize somehow (it can't sync to c.in anyways, since that is always waiting on a read in the RawMessage constructor).
    public void close() {
	closed = true;
	this.interrupt();
    }

    /**
     * forces this connection to close. Bad things will happen
     * if you use this carelessly.
     **/
    public void forceClose() {
	closed = true;
	c.close();
    }

    /**
     * Checks whether the connection is alive
     **/

    public boolean isOpen() {
	return isAlive() && !closed;
    }

    public Address peer() {
	return c.getPeerAddress();
    }

    public Address peer(ListeningAddress laddr) {
	return c.getPeerAddress(laddr);
    }

    public Address local() {
	return c.getMyAddress();
    }

    public Address local(ListeningAddress laddr) {
	return c.getMyAddress(laddr);
    }
}
