package Freenet;
import Freenet.support.*;
import Freenet.message.HandshakeRequest;
import java.util.*;
import java.net.*;
import java.io.*;

/*
  This code is part of the Java Adaptive Network Client by Ian Clarke. 
  It is distributed under the GNU Public Licence (GPL) version 2.  See
  http://www.gnu.org/ for further details of the GPL.
*/

/**
 * This is a Wrapper object that contains the components of a Node in the
 * Adaptive Network.  It uses a Network object to communicate with other
 * Nodes in the Network.
 *
 * @author <A HREF="mailto:I.Clarke@strs.co.uk">Ian Clarke</A>
 * @author <a href="mailto:blanu@uts.cc.utexas.edu">Brandon Wiley</a>
 **/

public class Core
{
    public static final String freenetVersion = "0.1";
    public static final String protocolVersion = "1.2";
    public static final String buildNumber = "117";
    public static String serverRevision;
    public static Params params;
    public static boolean debug;
    public static int timePerHop;
    public static int connectTimeout;
    public static int handshakeTimeout;
    public static int handshakeLife;
    public static boolean tunneling;

    public static Ticker timer;
    public ListeningAddress myAddress;
    public MessageHandler mh;
    public HandshakeHandler hh;
    public Listener listener;
    public boolean listen = true; // provides a way to turn off listening

    public Core(ListeningAddress myAddress, MessageHandler mh, HandshakeHandler hh) {
	try {
	    this.listener = ListenerFactory.listen(myAddress);
	} catch (BindException e) {
	    throw new RuntimeException("Port cannot be opened (illegal or in use)");
	}
	this.myAddress = listener.address;
	this.mh = mh;
	this.hh = hh;
	Logger.log("Core.java","Node running on "+listener,Logger.NORMAL);
    }
    
    public void acceptConnections()
    {
      Connection conn;
      RawMessage m;
      ConnectionHandler c;
      while(listen)
        {
          try
            {
              conn = listener.accept();
	      Logger.log("Node.java","Accepted connection:"+conn,Logger.MINOR);
            }
          catch (IOException e)
            {
              throw new RuntimeException("Problem accepting next connection");
            }
          c = new ConnectionHandler(conn, mh);
          c.start();
        }
    }
    

    /**
     * Makes a connection from this Node to another
     * @param peer The Address of the other node
     * @return A ConnectionHandler that handles the new connection
     **/

    public ConnectionHandler makeConnection(Address peer) throws ConnectFailedException {
	
	if(peer==null) throw new ConnectFailedException(peer);
	
	ConnectionHandler ch = connect(peer);

	if (!hh.getHandshake(this,ch))
	    throw new ConnectFailedException(peer);
    
	if (!ch.isOpen()) { // didn't stay open after handshake
	    ch = connect(peer);
	}

	return ch;
    }

    /**
     * This is the raw connection maker, it will not handshake
     * or check the status of the connection in any way.
     **/

    public ConnectionHandler connect(Address peer) throws ConnectFailedException {
	Connection c = null;
	try {
	    c = ConnectionFactory.connect(peer);
	} catch(Exception e) {
	    Logger.log("Node.java","Connection Error: " + e,Logger.ERROR);
	    throw new ConnectFailedException(peer);
	}

	Logger.log("Node.java","Connection between: " + c.getMyAddress(myAddress) + " and " + c.getPeerAddress(),Logger.DEBUGGING);
	if (c.getMyAddress(myAddress).equals(c.getPeerAddress())) {
	    Logger.log("Node.java","No connecting to yourself",Logger.MINOR);	    
	    c.close();
	    throw new ConnectFailedException(peer);
	}
	
	ConnectionHandler ch = new ConnectionHandler(c, mh);
	ch.start(); // start listening on this connection
	return ch;
    }

  /**
   * Send the message using the appropriate protocol
   * @depreciated Use makeConnection() to grab a ConnectionHandler and send from that instead.
   **/
    public void sendMessage(Message m, Address destination) throws SendFailedException
    {
	sendMessage(m, destination, null, null);
    }

    /**
     * Send the message using the appropriate protocol streaming
     * the actual data through the SplitOutputStream tunnel
     * @depreciated Use makeConnection to grab a ConnectionHandler and send from it instead
     **/

  public void sendMessage(Message m, Address destination, SplitOutputStream datatunnel) throws SendFailedException
    {
	sendMessage(m, destination, datatunnel, null);
    }

    /**
     * Send the message using the appropriate protocol, streaming
     * the actual data through the SplitOutputStream tunnel, and
     * counting the bytes sent with ByteCounter. This opens a connection as well.
     * @depreciated Use makeConnection to grab a ConnectionHandler and send from it instead
     **/

    public void sendMessage(Message m, Address destination, SplitOutputStream datatunnel, ByteCounter count) throws SendFailedException
    {
	try {
	    ConnectionHandler ch = makeConnection(destination);
	    ch.sendMessage(m, datatunnel, count);
	} catch (ConnectFailedException e) {
	    throw new SendFailedException(destination);
	}
    }
}
