package Freenet.client;
import java.io.*;
import java.net.*;
import java.util.*;
import Freenet.*;
import Freenet.crypt.*;
import Freenet.support.*;
import Freenet.message.*;

public class InsertClient
{
    private Address nodeAddress;
    private ListeningAddress myAddress;
    private long hopsToLive;
    private long startDepth;
    
    // option defaults
    static private int defaultListenPort = 0;
    static private String defaultServerAddress = "tcp/127.0.0.1:19114";
    static private long defaultHopsToLive = 5;
    static private int defaultLogging = Logger.NORMAL;
    static private int defaultVerbosity = 1;

    public ClientCore cn;

    // output all informational and error messages to this stream: (pixelpapst)
    static private PrintStream errout = System.err;


    static final private void keyinform(String key, String index)
    {
      errout.println("The keyindex server is only a temporary tool for alpha testers. It is not a permanent feature of the system. It will be removed in a future release.");
      try
      {
        (new URL(index+"?key="+key)).getContent();
      }
      catch (Exception e)
      {
          Logger.log("Node.java","Inform Connection failed:"+e, Logger.ERROR);
      }
    }

    static public void main(String[] args)
    {
	// parse command line
	Logger.threshold = Logger.ERROR;
	Params params = new Params(args);

	// check arguments
	if (params.getNumArgs() < 1 || params.getNumArgs() > 2) {
	    System.out.println("Usage: freenet_insert [options] key [input-file]");
	    System.out.println("");
	    System.out.println("  -length bytes              Bytes to insert (if reading stdin)");
	    System.out.println("  -listenPort port           Local port to listen for replies on");
	    System.out.println("  -serverAddress address     Server node to connect to");
	    System.out.println("  -htl hops-to-live          Hops-to-live for request");
	    System.out.println("  -logging error|normal|minor|debugging");
	    System.out.println("                             Logging level");
	    System.out.println("  -verbosity 1-5             Verbosity of log messages");
	    System.out.println("  -keyindex URL		     URL to notify that the key that is being inserted");
	    System.out.println("");
	    System.out.println("If input-file is omitted or is `-', read standard input.");
	    System.out.println("");
	    System.out.println("Examples:");
	    System.out.println("  freenet_insert freenet/manifesto manifesto.txt");
	    System.out.println("  cat manifesto.txt | freenet_insert -serverAddress tcp/localhost:19114 -htl 15 -length 1205 freenet/manifesto");
	    System.out.println("");
	    System.out.println("Send bug reports to freenet-dev@lists.sourceforge.net");
	    return;
	}

	// process options
	ListeningAddress lstaddr = new ListeningAddress
	    ("tcp/" + params.getint("listenPort", defaultListenPort));
	Address addr = new Address
	    (params.getParam("serverAddress", defaultServerAddress));
	SHA1 sha = new SHA1(true);
	Key key = new StringKey(sha.doHash(params.getArg(0)));
	long htl = params.getlong("htl", defaultHopsToLive);
        Logger.verbosity = params.getint("verbosity",defaultVerbosity);
        if (params.getParam("logging")==null)
            Logger.threshold = defaultLogging;
	else if (params.getParam("logging").equalsIgnoreCase("error"))
	    Logger.threshold = Logger.ERROR;
	else if (params.getParam("logging").equalsIgnoreCase("normal"))
	    Logger.threshold = Logger.NORMAL;
	else if (params.getParam("logging").equalsIgnoreCase("minor"))
	    Logger.threshold = Logger.MINOR;
	else if (params.getParam("logging").equalsIgnoreCase("debugging"))
	    Logger.threshold = Logger.DEBUGGING;

        String keyindex=params.getParam("keyindex");
        if(keyindex!=null)
          keyinform(params.getArg(0), keyindex);

	// open input file, if specified (default stdin)
	InputStream in = System.in;
	Long length = null;
	String inputFile = params.getArg(1);
	String inputLength = params.getParam("length");

	if (inputFile != null && !inputFile.equals ("-")) {
	    // reading from file
	    if (inputLength != null)
		errout.println ("freenet_insert: ignoring -length");

	    try {
		length = new Long(new File(inputFile).length());
		in = new FileInputStream(inputFile);
	    } catch (IOException ioe) {
		errout.println(ioe);
		System.exit(1);
		return;
	    }
	}
	else {
	    // reading from stdin
	    // make sure length was specified
	    if (inputLength == null) {
		errout.println ("freenet_insert: the -length option is mandatory when inserting from stdin");
		System.exit(1);
		return;
	    }
	    else {
		length = new Long(inputLength);
	    }
	}

        Core.connectTimeout=10000;
        Core.handshakeTimeout=10000;
	Core.handshakeLife=1000000;
        Core.timer=new Ticker(500);

	Logger.threshold=Logger.NORMAL;
   
	InsertClient ic = new InsertClient(addr, lstaddr, htl, (Math.abs((new Random()).nextLong()) % 30));
	
	try {
	    ic.insert(key, in, length, new DotCounter(length.intValue() / 80));
        }
        catch(BadHandshakeException e)
        {
          errout.println("Bad handshake");
	  System.exit(2);
          return;
	} catch (SendFailedException sfe) {
	    errout.print("Couldn't connect to the specified Freenet server - ");
	    if (params.getParam("serverAddress", defaultServerAddress).startsWith("tcp/localhost")) {
		errout.println("have you started it?");
	    }
	    else {
		errout.println("try a different one.");
	    }
	    System.exit(3);
       	    return;
	} catch (ClientException ce) {
	    if (ce.msg == null) {
		errout.println ("Network timed out waiting for a reply from the server.");
		System.exit(4);
		return;
	    } else if (ce.msg instanceof RequestFailed) {
		errout.println("The insert could not find a route. Lower the hops-to-live and try again."); 
	    } else if (ce.msg instanceof DataReply) {
		errout.println("Key collision, insert again under a different key.");
		synchronized(((DataReply)ce.msg).in) {
		    ((DataReply)ce.msg).in.notify();
		}
	    } else {
		errout.println("Received unexpected reply " + ce.msg);
	    }
	    System.exit(5);
	    return;
	}
    }

    public InsertClient(Address addr, ListeningAddress myAddr, long htl, long dpth)
    {
	nodeAddress = addr;
	hopsToLive = htl;
	startDepth = dpth;

	cn = new ClientCore(myAddr);

	myAddress = cn.myAddress;

    }

    public void insert(Key key, InputStream in, Long length, ByteCounter counter) throws ClientException, SendFailedException, BadHandshakeException
    {
	Random r = new Random(System.currentTimeMillis());

	ClientMessageHandler cmh = (ClientMessageHandler) cn.mh;
	cmh.ct = Thread.currentThread();

	ConnectionHandler ch;
	try {
	    ch = cn.connect(nodeAddress);
	} catch (ConnectFailedException e) {
	    throw new SendFailedException(e.peer);
	}

	HandshakeRequest hr = new HandshakeRequest(r.nextLong());
	hr.sending(cn, ch);

	ch.sendMessage(hr);

	try {
	    Thread.sleep(Core.handshakeTimeout);
	} catch (InterruptedException e) {
	}

	if (cmh.last == null || !(cmh.last instanceof HandshakeReply) || !((HandshakeReply) cmh.last).check())
	    throw new BadHandshakeException();
        errout.println("Kludgy handshake handling worked. "+cmh.last);
	
	long id = r.nextLong();

	InsertRequest ir=new InsertRequest(id, hopsToLive, startDepth, key);
	ir.sending(cn,ch);
	ch.sendMessage(ir);

	// clear last message
	cmh.last = null;

	// wait for InsertReply (or loop on QueryRestarted)
	boolean redo;
	do {
	    redo = false;
	    try {	
		Logger.log("client/InsertClient","Waiting another " + hopsToLive * 5 + " seconds for a response",Logger.NORMAL);
		Thread.sleep((hopsToLive * 5000) + 5000);
	    } catch (InterruptedException e) {
		if (cmh.last instanceof QueryRestarted) {
		    Logger.log("client/RequestClient","The request got stuck on a broken node but has been restarted at a depth of " + cmh.last.depth,Logger.MINOR);
		    redo = true;
		}
            }
	} while (redo);

	if (cmh.last == null || !(cmh.last instanceof InsertReply) || cmh.last.id != id) {
	    cn.listen = false;
	    ch.close();
	    throw new ClientException(cmh.last);
	} else {
	    DataInsert di = new DataInsert(id, hopsToLive, startDepth, nodeAddress);
    	    di.length = length;
	    di.in = in;
	    if (!ch.isOpen()) {
		try {
		    ch = cn.connect(nodeAddress);
		} catch (ConnectFailedException e) {
		    throw new SendFailedException(e.peer);
		}
	    }
	    di.sending(cn, ch);
	    ch.sendMessage(di, null, counter);
	    cn.listen = false;
	} 
    }
}
