package Freenet.client;
import java.net.*;
import java.io.*;
import java.util.*;
import Freenet.*;
import Freenet.crypt.*;
import Freenet.support.*;
import Freenet.message.*;

public class RequestClient
{
    private Address nodeAddress;
    private ListeningAddress myAddress;
    private long hopsToLive;
    private long startDepth;

    // option defaults
    static private int defaultListenPort = 0;
    static private String defaultServerAddress = "tcp/127.0.0.1:19114";
    static private long defaultHopsToLive = 5;
    static private int defaultLogging = Logger.NORMAL;
    static private int defaultVerbosity = 1;

    public ClientCore cn;
    public ConnectionHandler ch;
    public boolean closeOnExit = true;

    // output all informational and error messages to this stream: (pixelpapst)
    static private PrintStream errout = System.err;
    // this is important for piping to stdout


    static public void main(String[] args)
    {
	// parse command line
	Logger.threshold = Logger.ERROR;
	Params params = new Params(args);

        String keyindex=params.getParam("keyindex");
        if(keyindex!=null)
        {
          try
          {
            readKeys(keyindex);
            return;
          }
          catch(Exception e) {e.printStackTrace();}
        }

	// check arguments
	if (params.getNumArgs() < 1 || params.getNumArgs() > 2) {
	    System.out.println("Usage: freenet_request [options] key [output-file]");
	    System.out.println("");
	    System.out.println("  -listenPort port           Local port to listen for replies on");
	    System.out.println("  -serverAddress address     Server node to connect to");
	    System.out.println("  -htl hops-to-live          Hops-to-live for request");
	    System.out.println("  -logging error|normal|minor|debugging");
	    System.out.println("                             Logging level");
	    System.out.println("  -verbosity 1-5             Verbosity of log messages");
	    System.out.println("  -keyindex URL		     URL to fetch key list from");
	    System.out.println("");
	    System.out.println("If output-file is omitted or is `-', output on standard output.");
	    System.out.println("");
	    System.out.println("Examples:");
	    System.out.println("  freenet_request freenet/manifesto manifesto.txt");
	    System.out.println("  freenet_request -serverAddress tcp/localhost:19114 -htl 15 freenet/manifesto | more");
	    System.out.println("");
	    System.out.println("Send bug reports to freenet-dev@lists.sourceforge.net");
	    return;
	}

	// process options
	ListeningAddress lstaddr = new ListeningAddress
	    ("tcp/" + params.getint("listenPort", defaultListenPort));
	Address addr = new Address
	    (params.getParam("serverAddress", defaultServerAddress));
	SHA1 sha = new SHA1(true);
	Key key = new StringKey(sha.doHash(params.getArg(0)));
	long htl = params.getlong("htl", defaultHopsToLive);
        if (params.getParam("logging")==null)
            Logger.threshold = defaultLogging;
	else if (params.getParam("logging").equalsIgnoreCase("error"))
	    Logger.threshold = Logger.ERROR;
	else if (params.getParam("logging").equalsIgnoreCase("normal"))
	    Logger.threshold = Logger.NORMAL;
	else if (params.getParam("logging").equalsIgnoreCase("minor"))
	    Logger.threshold = Logger.MINOR;
	else if (params.getParam("logging").equalsIgnoreCase("debugging"))
	    Logger.threshold = Logger.DEBUGGING;
        Logger.verbosity = params.getint("verbosity",defaultVerbosity);

	// open output file, if specified (default stdout)
	OutputStream out = System.out;
	String outputFile = params.getArg(1);
	if (outputFile != null && !outputFile.equals ("-")) {
	    try {
		out = new FileOutputStream(outputFile);
	    } catch(IOException e) {
		errout.println(e);
		System.exit(1);
		return;
	    }
	}

	Core.connectTimeout = 30000;
        Core.handshakeTimeout = 30000;
	Core.handshakeLife=1000000;
        Core.connectTimeout=10000;
        Core.timer=new Ticker(500);

	Logger.threshold=Logger.NORMAL;

	RequestClient rc = new RequestClient(addr, lstaddr, htl, (Math.abs((new Random()).nextLong()) % 30));
	
	DataReply reply;
        InputStream stream;
	
	try {
	    reply = rc.request(key);
            stream=reply.in;
        } catch(BadHandshakeException e) {
            errout.println("Bad handshake from node");
	    System.exit(2);
	    return;
	} catch (SendFailedException sfe) {
	    errout.print("Couldn't connect to the specified Freenet server - ");
	    if (params.getParam("serverAddress", defaultServerAddress).startsWith("tcp/localhost")) {
		errout.println("have you started it?");
	    }
	    else {
		errout.println("please try a different one.");
	    }
	    System.exit(3);
	    return;
	} catch (ClientException ce) {
	    if (ce.msg == null) {
		errout.println("Network timed out waiting for a reply from the server.");
	    } else if (ce.msg instanceof RequestFailed) {
		errout.println("The key you requested could not be found anywhere in the network."); 
	    } else if (ce.msg instanceof TimedOut) {
		errout.println("The key you requested could not be found within HTL hops.");
	    } else {
		errout.println("Received unexpected reply: " + ce.msg);
	    }
	    System.exit(4);
	    return;
	}

	
	DotCounter counter=new DotCounter(reply.length.longValue() / 80);
	Conduit cond=new Conduit(stream, out, counter);

	cond.asyncFeed(null,null,reply.length.longValue());

    }

    private static final void readKeys(String index) throws Exception
    {
          System.out.println("The keyindex server is only a tool for alpha testers. It is not a permanent feature of the system. It will be removed in a future release.");
          InputStreamReader ir = new InputStreamReader((InputStream)(new URL(index).getContent()));
          BufferedReader br = new BufferedReader(ir);
          SHA1 sha=new SHA1(true);
          while(br.ready())
          {
            System.out.println(br.readLine());
          }
    }

    public RequestClient(Address addr, ListeningAddress myAddr, long htl, long dpth)
    {
	nodeAddress = addr;
	hopsToLive = htl;
	startDepth = dpth;
	cn = new ClientCore(myAddr);
	myAddress = cn.myAddress;

    }

    public DataReply request(Key key) throws ClientException, SendFailedException, BadHandshakeException
    {
	Random r = new Random(System.currentTimeMillis());

	ClientMessageHandler cmh = (ClientMessageHandler) cn.mh;
	cmh.ct = Thread.currentThread();

	try {
	    ch = cn.connect(nodeAddress);
	} catch (ConnectFailedException e) {
	    throw new SendFailedException(e.peer);
	}

	HandshakeRequest hr = new HandshakeRequest(r.nextLong());
	hr.sending(cn, ch);
	ch.sendMessage(hr);

	try {
	    Thread.sleep(10000);
	} catch (InterruptedException e) {
	}

	if (cmh.last == null || !(cmh.last instanceof HandshakeReply) || !((HandshakeReply) cmh.last).check())
	    throw new BadHandshakeException();

	DataRequest dr=new DataRequest(r.nextLong(), hopsToLive, startDepth , key);
	dr.sending(cn, ch);
	ch.sendMessage(dr);

	// clear last message
	cmh.last = null;
	
	// wait for InsertReply (or loop on QueryRestarted)
	boolean redo;
	do {
	    redo = false;
	    try {
		Logger.log("client/RequestClient","Waiting another " + hopsToLive * 5 + " seconds for a response",Logger.NORMAL);
		Thread.sleep(hopsToLive * 5000 + 5000);
	    } catch(InterruptedException e) {
		if (cmh.last instanceof QueryRestarted) {
		    Logger.log("client/RequestClient","The request got stuck on a broken node but has been restarted at a depth of " + cmh.last.depth,Logger.MINOR);
		    redo = true;
		}
	    }
	} while (redo);

	if (cmh.last == null || !(cmh.last instanceof DataReply) || cmh.last.id != dr.id) {
		if (closeOnExit) {
		    Logger.log("client/RequestClient","Closing connection",Logger.NORMAL);
		    ch.close();
		}
	    throw new ClientException(cmh.last);
	} else {
	    return (DataReply)cmh.last;
	}
    }

	public void setCloseOnExit (boolean closeOnExit) {
		this.closeOnExit = closeOnExit;
	}
}
