#!/usr/bin/perl -w
# Liberator v 0.03
# Freenet client by Oskar Sandberg
#
# see http://freenet.sourceforge.net for more info.
#
# This code is released under the GNU Public License. I hate legal babel.
# Warranty my ass.
#
# version history
#
# 991228 v 0.001
# 2K0102 v 0.002 Removed fork (thanks Steven Hazel) and fixed 
#                handling of FailedRequest (Michael fixed them in 
#                Freenet).
# 2K0116 v 0.03  Now asks socket for local address, can save/read from
#                a file, supports freenet protocol changes, checks
#                for correct id (and server) on reply, added verbous, strict, 
#                dh, and dl options, and cleaned up syntax.
#
# 2K0304 v 0.04  Change for new handshake protocol using message format
#                Update FREENETPROTO for this new protocol
#                Use SHA hash of key
#                Include DataLength field in insert messages
#                Change default port to be consistent with current distribution
#                (by Hal Finney)
# 2K0305 v 0.05  Change to use native SHA-1 rather than external module
# 2K0306 v 0.06  Change to use new two-message Insert protocol
# 2K0309 v 0.07  Use new two-message request/reply Handshake protocol
# 2K0320 v 0.08  Add EndMessage field; handle QueryRestarted message
# 2K0328 v 0.09  Support for sustained connections between messages,
#                protocol 1.2.
# 2K0408 v 0.10  Silent option and progress indicator + some fixes

use strict;
use Socket;
use Getopt::Long;
use FileHandle;

my ($remote, $port, $iaddr, $paddr, $paddr2, $proto, $line, $message, $id, $depth, $handshake, $pid, $reply);

my $FREENETPROTO = "1.2";
my $BUFFER = "256";
my $DEFAULTNODE = "localhost";
my $DEFAULTPORT = 19114;
my $DEFAULTHOPS = 10;
my $DEFAULTDL = 5;
my $DEFAULTDH = 30;

sub usage {
  print(" " x 10 . "Liberator v 0.08, Freenet Client by Oskar Sandberg\n",
	" " x 18 .  "The supported protocol is $FREENETPROTO\n",
	"THIS SOFTWARE COMES WITH ABSOLUTELY NO WARRANTY! USE AT YOUR OWN RISK!\n\n",
	"usage: liberator -i|--insert -k|--key=KEY [Options]\n",
	"   or: liberator -r|--request -k|--key=KEY [Options]\n",
	"Options:\n",
	"-n, --node=ADDRESS ($DEFAULTNODE\:$DEFAULTPORT)   Address of first Freenet node\n",
	"-h, --hops=HOPSTOLIVE ($DEFAULTHOPS)   Messages hops to live on Freenet\n",
	"-f, --file=FILENAME (stdout/stdin)   Filename to write to or read from\n",
	"-dl, --depthlow=LOWDEPTH ($DEFAULTDL)\n",
	"-dh,--depthhigh=HIGHDEPTH ($DEFAULTDH)\n   The original depth of the message is set to random number between\n   depthhigh and depthlow. The former must be smaller or equal (duh)\n",
	"--verbose   Print sent message to stdout (can't be used with --silent)\n",
	"--strict   Use strict security\n",
        "--silent   Silence all output (can't be used with --verbose)\n",
        "--help   Print this screen\n",
	"\nFor more info see The Freenet Project at http:\\freenet.sourceforge.net\n");
  exit 0;
}

my ($insert, $request, $node, $key, $hop, $dl, $dh, $verbose, $filename, $strict, $help, $silent);
GetOptions("insert|i" => \$insert,
	   "request|r" => \$request,
	   "node|n=s" => \$node,
	   "key|k=s" => \$key,
	   "hop|h=i" => \$hop,
	   "dl|depthlow=i" => \$dl,
	   "dh|depthhigh=i" => \$dh,
	   "verbose" => \$verbose,
	   "file|f=s" => \$filename,
	   "help" => \$help,
	   "strict" => \$strict,
	   "silent" => \$silent) || usage;

# Check that all the paras are set correct.

$strict && ($dh && $dl || die("STRICT: depthhigh and depthlow must be set (never trust a machine)\n"));

(($insert xor $request) && $key && ($hop || ($hop = $DEFAULTHOPS)) && ($dl || ($dl = $DEFAULTDL)) && ($dh || ($dh = $DEFAULTDH)) && ($dh >= $dl) && !$help) &&!($silent && $verbose) && ($filename || ($silent = 1)) || usage;

($remote, $port) = split(/\:/,$node) if $node;

($port || ($port = $DEFAULTPORT)) && ($remote || ($remote = $DEFAULTNODE));

if ($port =~ /\D/) { $port = getservbyname($port, 'tcp') }
usage unless $port;
$iaddr   = inet_aton($remote) || die "no host: $remote";
$paddr   = pack_sockaddr_in($port, $iaddr);
$proto   = getprotobyname('tcp');

# now open the server socket to send the DataInsert or DataRequest
socket(SOCK, PF_INET, SOCK_STREAM, $proto)  || die "socket: $!";
connect(SOCK, $paddr) || die "connect: $!";
# autoflush
SOCK->autoflush(1);

# Generate UniqueID
$id = '';
# use /dev/random if possible
if (open(RAND,"/dev/random")) {
  read RAND, $id, 8;
  $id = pack("a16",unpack("h16",$id));
}
# use /dev/urandom if that didn't work
if (length($id) < 16 && open(URAND,"/dev/urandom")) {
  read URAND, $id, 8;
  $id = pack("a16",unpack("h16",$id));
}
# else use rand() if not strict
if (length($id) < 16) {
  if ($strict) {
    die("Failed to created random UniqueID for message");
  } else {
    for (length($id)+1..16) {$id .= sprintf("%x",int(rand(16)))}
  }
}

substr($id,0,1) = '1' if substr($id,0,1) =~ "0";

# assemble handshake message

  $message = "HandshakeRequest\n";
  $message .= "UniqueID=$id\n";
  $message .= "Depth=1\n";
  $message .= "HopsToLive=1\n";
  $message .= "EndMessage\n";
  
  print SOCK $message;
  print $message if $verbose;

# Handle handshake reply
$reply = &handlereply (\*SOCK);
die("Did not get handshake back\n") unless $reply =~ /HandshakeReply/;

# assemble next message

if ($insert) {
  $message = "InsertRequest\n";
} elsif ($request) {
  $message = "DataRequest\n";
}

# $message .= "Source=tcp/$myhost:0\n";

$key = sha1hex($key);
$message .= "SearchKey=$key\n";


# Change $id to be different
if(substr($id,0,1) =~ "1") {
  substr($id,0,1) = '2';
} else {
  substr($id,0,1) = '1';
}
$message .= "UniqueID=$id\n";
$message .= "Depth=" . (int(rand($dh - $dl + 1)) + $dl) . "\n";
$message .= "HopsToLive=$hop\n";
$message .= "EndMessage\n";



print SOCK $message;
print $message if $verbose;

# Print the reply
$reply = &handlereply(\*SOCK);

# Get second message for insert
if ($insert) {
  die("Could not send data\n") unless $reply =~ /InsertReply/;

  my ($infile, @indata, $count, $time);

  $message = "DataInsert\n";
  $message .= "DataSource=tcp/$remote:$port\n"; # lie about datasource
  $message .= "SearchKey=$key\n";
  $message .= "UniqueID=$id\n";
  $message .= "Depth=" . (int(rand($dh - $dl + 1)) + $dl) . "\n";
  $message .= "HopsToLive=$hop\n";

  print SOCK $message;
  print $message if $verbose;

  if ($filename) {
    open(INFILE, $filename) || die("Could not open $filename");
    $count = (stat(INFILE))[7];
    print SOCK "DataLength=$count\n";
    print "DataLength=$count\n" if $verbose;
    print SOCK "Data\n";
    print "Data\n" if $verbose;
    my($read, $buffer, $ind);
    $read = 0; $ind = 0; 
    print "Sending data\n" unless $silent;
    $time = time;
    do {
      $read += read INFILE, $buffer, ($BUFFER > ($count - $read) ? $count - $read : $BUFFER);

      print SOCK $buffer;
      if (!$silent && ($read > $ind)) {
	print ".";
	$ind += $count/80;
      }
    } while ($read < $count);
    $time = time - $time;
    print "\nSent $count bytes in $time seconds.\n" unless $silent;
  } else {
    #  read all of data in so we can get its length
    @indata = <STDIN>;
    
    $count = 0;
    foreach $line ( 0 .. $#indata ) {
      $count += length($indata[$line]);
    }
    print SOCK "DataLength=$count\n";
    print "DataLength=$count\n" if $verbose;
    print SOCK "Data\n";
    print "Data\n" if $verbose;
    print SOCK @indata;
  }
} elsif ($request) {
  die("The data was not found\n") unless $reply =~ /DataReply/;
}

close SOCK;
exit;

# Check for valid handshake and version
sub handlereply {
  my $socket = shift;
    
    print "_________ \n Received:\n" if $verbose;
    my $messagetype = <$socket>;
    print ("$messagetype\n") if $verbose;
    my(%fields, $field);      # read field values
    while (defined($field = <$socket>) && ($field  =~ /\=/)) {
      
      print $field if $verbose;
      my($fieldname, $fieldvalue) = split('=',$field);
      chop $fieldvalue;      #dump \n at the end of every field
      $fields{$fieldname} = $fieldvalue;
    };
    print "\n" if $verbose;
  
    if ($fields{'UniqueID'} ne $id  && !($messagetype =~ /HandshakeRequest/) ) {
      print "Recieved message with wrong id!\n" unless $silent;
      print "Recieved message '$fields{'UniqueID'}' expected '$id'\n" if ($verbose);
      return 0;
    }
    if ($messagetype =~ /HandshakeRequest/) {

      $message = "HandshakeReply\n";
      $message .= "Depth=$fields{'Depth'}\n";
      $message .= "HopsToLive=$fields{'HopsToLive'}\n";
      $message .= "UniqueID=$fields{'UniqueID'}\n";
      $message .= "Version=$FREENETPROTO\n";
      $message .= "EndMessage\n";
      print $socket $message;
    
      print "Sending handshake reply message:\n" if $verbose;
      print $message if $verbose;

      #recursive call to listen for another message (too much lisp)
      return &handlereply($socket);

    } elsif ($messagetype =~ /DataReply/) {
      if ($request) { 
  	my($outfile,$read,$buffer,$ind, $time); 
	
  	if ($filename) {
  	  die("File '$filename' already exists\n") if (-f $filename);
  	  open(OUTFILE, ">$filename") || die("Could not open $filename");
  	  $outfile = \*OUTFILE;
  	} else {
  	  $outfile = \*STDOUT;
  	}
	$read = 0; $ind = 0;
	print "Retrieving Data\n" unless $silent;
	$time = time;
  	do {
	  $read += read $socket, $buffer, ($BUFFER > ($fields{"DataLength"} - $read) ? $fields{"DataLength"} - $read : $BUFFER);
	  if (!$silent && ($read > $ind)) {
	    print ".";
	    $ind += $fields{"DataLength"}/ 80;
	  }
  	  print $outfile $buffer;
  	} while ($read < $fields{"DataLength"});
	$time = time - $time;
	print "\nRetrieved " . $fields{"DataLength"} . " bytes in $time seconds.\n" unless $silent;

      } elsif($insert) {
  	print "You insert collided with previous data with the same key. \n" .
	  "Please insert it again with a new key and discourage anyone\n" .
	  "from attempting to access your data with this key. We recommend\n" .
	  "you attempt to request the key before inserting to make sure it\n" .
          "is not used already\n" unless $silent;
      }
    } elsif ($messagetype =~ /HandshakeReply/) {
      die "Server running version $fields{'Version'}, we support $FREENETPROTO"
           unless( $fields{'Version'} == "$FREENETPROTO" );
    } elsif ($messagetype =~ /InsertReply/) {
      ;
    } elsif ($messagetype =~ /QueryRestarted/) {
      #wait more
      return &handlereply($socket);
    } elsif ($messagetype =~ /RequestFailed/) {
      print "The entire network was searched, you are probably cut of Freenet (or your Hops To Live is way to high)\n" unless $silent;
     } elsif ($messagetype =~ /TimedOut/) {
      if ($request) {
  	print "Your request timed out, if you are sure the data exists try raising the Hops To Live (-h)\n" unless $silent;
      } elsif ($insert) {
  	print "You insert collided with a reference for which data could\nnot be found. Use a different key or try waiting out the old references\n" unless $silent;
      }
    } else {
      print "Recieved unknown reply. Maybe an upgrade is in order" unless $silent;
    }
  return $messagetype;
}

# SHA-1 in Perl
sub sha1hex {
  my ( $instr ) = @_;
  my ( @A, @K, @F, @W, $total, $bytes, $padded, $t, $a, $b, $c, $d, $e, $r, $r20 );

  # Initial values
  @A = ( 0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476, 0xc3d2e1f0 );
  # Constants used in algorithm
  @K = ( 0x5a827999, 0x6ed9eba1, 0x8f1bbcdc, 0xca62c1d6 );
  # Four different nonlinear functions used in different iterations
  @F = (sub{$b&($c^$d)^$d},sub{$b^$c^$d},sub{($b|$c)&$d|$b&$c},sub{$b^$c^$d});
  # Fix overflowing adds to be mod 2^32
  sub addfix{
   my( $val ) = @_;
   my( $denom );
   $denom = 1 + ~0;
   return $val - ($denom * int( $val/$denom ));
   };
  # Rotate left the specified amount
  sub lrotate{
   my( $val, $rot ) = @_;
   my( $mask );
   $mask = 2**$rot - 1;
   return ($val<<$rot) | (($val>>(32-$rot)) & $mask);
   }

  # Loop for each 512 bit block of input
  do {
   # Get next 512 bits of input
   $_ = substr($instr, 0, 64);
   substr($instr, 0, 64) = "";
   $total += $bytes = length $_;

   # Add byte of 0x80 after last byte of input
   if( $bytes < 64 && !$padded ) {
     $bytes++;
     $_ .= "\x80";
     $padded = 1;
   }

   # Convert 64 bytes of input into 16 words
   @W = unpack "N16", $_ . "\0"x64;

   # If last block, set total bit length in last word.
   if ($bytes <= 56) {
     $W[15] = $total*8;
   }

   # Fill rest of word array.  Note that lrotate distinguishes SHA-1 from SHA.
   for $r (16..79){
    $W[$r] = lrotate($W[$r-3]^$W[$r-8]^$W[$r-14]^$W[$r-16], 1)
   }

   # Get state variables $a - $e.
   ($a,$b,$c,$d,$e) = @A;

   # Loop for 80 rounds
   for $r (0..79){
    # Do one round of SHA transformation
    $r20 = $r / 20;
    $t = addfix( lrotate($a, 5) + &{$F[$r20]} + $e + $W[$r] + $K[$r20] );
    $e = $d;
    $d = $c;
    $c = lrotate($b, 30);
    $b = $a;
    $a = $t
   }

   # Add back into initial state variables to finish processing this block
   $A[0] = addfix($A[0]+$a);
   $A[1] = addfix($A[1]+$b);
   $A[2] = addfix($A[2]+$c);
   $A[3] = addfix($A[3]+$d);
   $A[4] = addfix($A[4]+$e);
  } while ($bytes > 56);

  # Done, return in Freenet format, 8 upper case chars followed by space.
  return sprintf( "%.8X"x5, @A );
}
