package Freenet.contrib.fproxy;
import Freenet.*;
import Freenet.client.*;
import Freenet.support.*;
import java.io.*;
import java.net.*;
import java.util.*;

/*
  This code is part of fproxy, an HTTP proxy server for Freenet.
  It is distributed under the GNU Public Licence (GPL) version 2.  See
  http://www.gnu.org/ for further details of the GPL.
*/


/**
 * Thread to handle incoming HTTP connections for the ProxyServer
 *
 * @author <a href="http://www.doc.ic.ac.uk/~twh1/">Theodore Hong</a>
 **/

public class HttpHandler extends Thread
{
    // local variables
    protected static final char[] lf = {'\012'};

    private ProxyServer ps;
    private Socket s;
    private InputStream in;
    private OutputStream out;
    private static long ids=0;
    private long id;


    /**
     * Creates a new HttpHandler communicating on Socket s for ProxyServer ps
     **/
    public HttpHandler(Socket s, ProxyServer ps) throws IOException {
	this.ps = ps;
	this.s = s;
	this.in = s.getInputStream();
	this.out = s.getOutputStream();
	this.id = ++ids;

	Logger.log("HttpHandler.java#"+id, "New HttpHandler with "+s.getInetAddress(), Logger.NORMAL);
    }


    /**
     * Main routine to process an HTTP request
     **/
    public void run() {
	try {
	    // wait for request to arrive
	    while (in.available() == 0) {
		try {
		    Thread.sleep(50);
		} catch (InterruptedException e) {}
	    }

	    // read key
	    PushbackInputStream pstream = new PushbackInputStream(in);
	    String line = readTo(pstream, lf);
	    StringTokenizer st = new StringTokenizer (line, " ");
	    st.nextToken();
	    String key = st.nextToken();

	    // strip extraneous prefixes
	    String[] prefixes = {"http://", "/"};
	    for (int x = 0; x < prefixes.length; x++) {
		if (key.startsWith(prefixes[x])) {
		    key = key.substring(prefixes[x].length());
		    break;
		}
	    }

	    // send response header
	    Logger.log("HttpHandler.java#"+id, "Sending HTTP response header", Logger.NORMAL);
	    PrintWriter pw = new PrintWriter (out);
	    pw.print ("HTTP/1.0 200 OK\015\012");
	    pw.print ("Connection: close\015\012");
	    // for now, we'll send this instead of ProxyClient
	    pw.print("Content-type: text/plain\015\012\015\012");
	    pw.flush();

	    Logger.log("HttpHandler.java#"+id, "Transferring log messages to HTTP stream", Logger.NORMAL);
	    Logger.logto(out);

	    /*
	      HTTP/1.1 200 OK
	      Date: Wed, 05 Apr 2000 00:56:37 GMT
	      Server: Apache/1.3.12 (Unix)
	      Last-Modified: Mon, 03 Apr 2000 12:20:42 GMT
	      ETag: "f7a42-36cc-38e88c9a"
	      Accept-Ranges: bytes
	      Content-Length: 14028
	      Connection: close
	      Content-Type: text/html
	    */

	    // do freenet request
	    // note: ProxyClient will eventually send Content-type and terminate the headers
	    Logger.log("HttpHandler.java#"+id, "Requesting key: "+key+"\n", Logger.NORMAL);
	    ProxyClient pc = new ProxyClient (0, ps.serverAddress, ps.hopsToLive, out);
	    pc.request (key);

	    // wait for request to complete
	    Logger.logto(ps.Lout);
	    Logger.log("HttpHandler.java#"+id, "Waiting for request to complete", Logger.NORMAL);
	    if (pc.rc.ch != null) {
		try {
		    // join ProxyClient.RequestClient.ConnectionHandler thread
		    pc.rc.ch.join();
		}
		catch (InterruptedException e) {}
	    }

	    // cleanup
	    Logger.log("HttpHandler.java#"+id, "Request done", Logger.NORMAL);
	    in.close();
	    out.close();
	    s.close();

	    Logger.log("HttpHandler.java#"+id, "Connection closed", Logger.DEBUGGING);

	// just exit if there are any io/socket errors
	} catch (EOFException e) {
	} catch (IOException e) {
	}
    }
    

    /**
     * Utility function to return String from InputStream until one of
     * the characters in tms is about to be read.
     **/
    protected String readTo(PushbackInputStream i, char[] tms)
	throws IOException, EOFException
    {
	StringBuffer tmp = new StringBuffer();
	char r=' ';
	int ir;
	while (true) {
	    ir = i.read();
	    if (ir == -1)
		throw new EOFException();
	    r = (char) ir;
	    if (r == -1)
		throw new EOFException();
	    boolean b = false;
	    for (int x=0; x<tms.length; x++) {
		if (tms[x] == r) {
		    b=true;
		    break;
		}
	    }
	    if (b)
		break;
	    tmp.append(r);
	}
	if (ir != -1)
	    i.unread((int) r);
	return tmp.toString();
    }
}
