package Freenet.contrib.fproxy;
import Freenet.*;
import Freenet.support.*;
import java.io.*;
import java.net.*;


/*
  This code is part of fproxy, an HTTP proxy server for Freenet.
  It is distributed under the GNU Public Licence (GPL) version 2.  See
  http://www.gnu.org/ for further details of the GPL.
*/


/**
 * The main ProxyServer class.  It listens for incoming HTTP connections on 
 * -listenPort (default 80), extracts the Freenet key, and issues a proxy 
 * request on the web client's behalf to the Freenet node at -serverAddress 
 * (default tcp/localhost:19114).  The reply is streamed directly over the
 * HTTP connection back to the web client, with the appropriate MIME type
 * (currently, always text/plain).  Any errors are also reported to the 
 * web client.
 *
 * @author <a href="http://www.doc.ic.ac.uk/~twh1/">Theodore Hong</a>
 **/

public class ProxyServer
{
    // server parameters
    public int listenPort;
    public String serverAddress;
    public long hopsToLive;
    public PrintStream Lout = Logger.Lout;

    // defaults
    static private int defaultListenPort = 80;
    static private String defaultServerAddress = "tcp/localhost:19114";
    static private long defaultHopsToLive = 5;
    static private int defaultLogging = Logger.NORMAL;
    static private int defaultVerbosity = 1;

    static public void main(String[] args) {
	// parse command line
	Logger.threshold = Logger.ERROR;
	Params params = new Params(args);

	// check for help
	if (params.getParam("help") != null) {
	    usage();
	    return;
	}

	// process options
	int port = params.getint("listenPort", defaultListenPort);
	String addr = params.getParam("serverAddress", defaultServerAddress);
	long htl = params.getlong("htl", defaultHopsToLive);
	ProxyServer ps = new ProxyServer(port, addr, htl);

	// set logging
        if (params.getParam("logging")==null)
            Logger.threshold = defaultLogging;
	else if (params.getParam("logging").equalsIgnoreCase("error"))
	    Logger.threshold = Logger.ERROR;
	else if (params.getParam("logging").equalsIgnoreCase("normal"))
	    Logger.threshold = Logger.NORMAL;
	else if (params.getParam("logging").equalsIgnoreCase("minor"))
	    Logger.threshold = Logger.MINOR;
	else if (params.getParam("logging").equalsIgnoreCase("debugging"))
	    Logger.threshold = Logger.DEBUGGING;
        Logger.verbosity = params.getint("verbosity",defaultVerbosity);

	String fname = params.getParam("logFile","NO");
	if (!fname.equalsIgnoreCase("NO")) {
	    try {
		Logger.logto(fname);
	    } catch (Exception e) {
		System.out.println("Writing to log failed");
	    }
	}
	ps.Lout = Logger.Lout;
	
	// bind local port
	ServerSocket sock;
	try {
	    Logger.log("ProxyServer.java", "Proxy server started on port "+port, Logger.NORMAL);
	    sock = new ServerSocket(port);
	}
	catch (IOException e) {
	    Logger.log("ProxyServer.java", "Error binding local port: "+e, Logger.ERROR);
	    return;
	}

	// listen for connections
	while(true) {
	    try {
		Socket a = sock.accept();
		Logger.log("ProxyServer.java", "Got new connection: "+a,
			   Logger.MINOR);
		HttpHandler h = new HttpHandler(a, ps);
		h.start();
            }
	    catch (IOException e) {
		throw new RuntimeException("Error accepting next connection:"+e);
            }
        }
    }


    /**
     * Constructor which stores the invocation parameters of this ProxyServer.
     **/
    public ProxyServer(int port,String addr, long htl) {
	this.listenPort = port;
	this.serverAddress = addr;
	this.hopsToLive = htl;
    }


    /**
     * Print usage information
     **/
    public static void usage() {
	System.out.println("fproxy version 0.1");
	System.out.println("Usage: fproxy [options]");
	System.out.println("");
	System.out.println("  -listenPort port           Local port to listen on");
	System.out.println("  -serverAddress address     Server node to connect to");
	System.out.println("  -htl hops-to-live          Hops-to-live for request");
	System.out.println("  -logFile file              Name of log file (`no' to log to console)");
	System.out.println("  -logging error|normal|minor|debugging");
	System.out.println("                             Logging level");

	System.out.println("  -help                      Display this help and exit");
	System.out.println("");
	System.out.println("Send bug reports to Theodore Hong <twh1@doc.ic.ac.uk>.");
    }
}

