package Freenet.message;
import java.io.*;
import Freenet.*;
import Freenet.node.*;
import Freenet.support.*;

/*
  This code is part of the Java Adaptive Network Client by Ian Clarke. 
  It is distributed under the GNU Public Licence (GPL) version 2.  See
  http://www.gnu.org/ for further details of the GPL.

  Explanation of Code Versions: 
    0.0.0      = Initial Description
    0.0.1      = API Specified
    0.x (x>0)  = Partial Implementation
    x.0 (x>0)  = Operational
		
  Requires Classes: Node (1.0)
                    Address (1.0)
		    Message (1.0)
 */

/**
 * This is the DataReply message
 *
 * @see Node
 * @see Address
 * @author Brandon Wiley (blanu@uts.cc.utexas.edu)
 * @author Ian Clarke (I.Clarke@strs.co.uk)
 **/

public abstract class DataSend extends Message
{
  public InputStream in=null;
  public Long length=null;
  public Address dataSource=null;

  public DataSend(long idnum, long htl, long depth, Address dsrc)
    {
      super(idnum, htl, depth);
      this.dataSource = dsrc;
    }

    public DataSend(long idnum, long htl, long depth, Address dsrc, Data data) throws DataNotReadyException, IOException {
	super(idnum, htl, depth);
	this.dataSource = dsrc;
	length=data.getLength();
	in=data.toInputStream();
    }

  public DataSend(RawMessage raw) throws InvalidMessageException
  {
    super(raw);
    dataSource = new Address(popField("DataSource"));
    if(raw.trailingFieldLength!=null)
      length=raw.trailingFieldLength;
    else
      throw new InvalidMessageException("Data sending message requires the trailing field length to be specified");
    if(raw.trailingFieldName!=null)
	in = raw.trailingFieldStream;
 
    if(in==null) throw new InvalidMessageException("Data sending message requires a trailing field called Data");
  }

  public RawMessage toRawMessage()
  {
    RawMessage raw=super.toRawMessage();
    raw.setField("DataSource", dataSource.toString());
    if(in!=null)
    {
	raw.trailingFieldStream=in;
        raw.trailingFieldLength=length;
    }
    else
      return raw;
    raw.trailingFieldName="Data";
    return raw;
  }

  /**
   * Called by a node after it receives this message
   * @param n The node that called this message.  This should be used
   *          to make any nescessary modifications to the DataStore etc.
   * @param sb Null unless this node has been seen before.  If non-null it
   *           is the Object returned by the received method of the
   *           last message seen with this ID.
   * @return The object to be passed to any messages received in the near
   *         future with the same ID as this one.  Null if no object should
   *         be passed.
   **/
  public MessageMemory pReceived(Node n, MessageMemory sb)
  {
      Logger.log("message/DataSend.java","Got a DataSend!",Logger.MINOR);
    if(sb==null || !(sb instanceof KeyedMM)) {// No one requested this, or I forgot about it
	//	try{in.close();} catch(IOException e) {}
	return null;
    }
    else // Forward it to whoever requested it.
    {
	// Cancel any timer waiting for this message
	Node.timer.cancel(id);

	SplitOutputStream datatunnel = new SplitOutputStream();
	KeyedMM kmm = (KeyedMM)sb;
	if (n.ds.searchData(kmm.searchKey) == null) { 
	    // if I got some data between request and reply, don't recache
	    n.ds.put(kmm.searchKey, dataSource, new Data(datatunnel)); // Cache the data
	}

	if (kmm.dataref == null && kmm.replyCon == null) { // So weeeeee've come, to the eeeeeend of the road...
	    Logger.log("message/DataSend.java","Message ends here",Logger.DEBUGGING);
	    (new Conduit(in, datatunnel)).asyncFeed(datatunnel,null,length.longValue());
	} else { // still got places to go
	    try {
		if (kmm.replyCon == null || !(sb.replyCon.isOpen())) {
		    sb.replyCon = n.makeConnection(kmm.dataref);
		}
		sending(n, sb.replyCon);
		sb.replyCon.sendMessage(this, datatunnel, null);
	    } catch(ConnectFailedException cfe) {
		Logger.log("message/DataSend.java","Couldn't restore connection to proxy data to " + cfe.peer,Logger.NORMAL);
	    } catch(SendFailedException sfe) {
		Logger.log("message/DataSend.java","Couldn't proxy data to " + sfe.peer,Logger.NORMAL);
	    } 
	}

	kmm.dataref = null;
	kmm.replyCon = null;
	kmm.pending = false;
	return kmm;
       
	//	return null; // My job is done. Forget about this.
    }
  }

    public void sending(Core n, Address peer, Address myAddress) throws SendFailedException
    {
	if (myAddress.equals(peer))
	    throw new SendFailedException(peer); // don't send to yourself
	source = myAddress;
    }

      // Protected/Private Methods

    protected MessageMemory timeOut(Node n, MessageMemory sb)
    {
	SplitOutputStream datatunnel = new SplitOutputStream();	

	if (sb != null && sb instanceof KeyedMM && n.ds.searchData(((KeyedMM) sb).searchKey) == null) {
	    KeyedMM kmm = (KeyedMM) sb; 
	    Data mydata = new Data(datatunnel);
	    n.ds.put(kmm.searchKey, dataSource, mydata);
	} else {
	    Logger.log("DataSend.java","Not saving data",Logger.DEBUGGING);
	}

	(new Conduit(in, datatunnel)).asyncFeed(datatunnel,null,length.longValue());

	return null;
    }
    
}


