/*
 *$Id: StandardDataStore.java,v 1.10 2000/04/08 20:48:17 hobbex Exp $

  This code is part of the Java Adaptive Network Client by Ian Clarke.
  It is distributed under the GNU Public Licence (GPL) version 2.  See
  http://www.gnu.org/ for further details of the GPL.

  Explanation of Code Versions:
    0.0.0      = Initial Description
    0.0.1      = API Specified
    0.x (x>0)  = Partial Implementation
    x.0 (x>0)  = Operational

  Requires Classes: Address
                    Data
 */

/**
 * DataStore.java
 *
 * This class is used by a node to store references and data.
 *
 * @version 0.0
 * @author Ian Clarke (I.Clarke@strs.co.uk)
 **/

package Freenet.node;
import Freenet.*;
import Freenet.support.*;
import java.io.*;
import java.util.*;

public class StandardDataStore implements DataStore
{
     static private long write_interval;
     static public String filename = ".freenet/store";

  // Debugging stuff to test this class
    /* not working since Ian removed test code from data class
  public static void main(String[] args) {
    System.out.println("DataStore test routine\n-------");
    System.out.println("Creating datastore size 20 with room for 30 bytes of data");
    DataStore ds = new DataStore(30, 20);
    System.out.println("Adding 10 data items");
    for (int x = 0; x < 100; x += 5)
      ds.put(new StringKey("Key "+x), new Address("local/"+x),
             new Data("Data "+x));
    System.out.println("Adding 10 reference items");
    for (int x = 100; x < 200; x += 5)
      ds.put(new StringKey("Key "+x), new Address("local/"+x), null);
    Key last = null;
    for (int x = 0; x < 10; x++) {
      System.out.println("Finding closest key to Key 138 with mask "+
                         last);
      last = ds.findClosestKey(new StringKey("Key 138"), last);
      System.out.println("Found "+last);
      }
    }
    */
  public static DataStore makeDataStore(Node node,  long maxData, int maxItems)
  {
    Object o=null;

    try
    {
      System.out.println(filename);
      FileInputStream stream=new FileInputStream(filename);
      ObjectInputStream ostream=new ObjectInputStream(stream);
      o=ostream.readObject();
      ostream.close();
      if(!(o instanceof DataStore))
        o=new StandardDataStore(maxData, maxItems);
    }
    catch(Exception e) 
	{
	    Logger.log("DataStore.java", 
		       "Couldn't load DataStore from "+filename, 
		       Logger.NORMAL);
	    o=new StandardDataStore(maxData, maxItems); 
	}

    write_interval=1000*Node.params.getlong("checkpointInterval",5*60);
    Node.timer.add(23, write_interval, (Callback)o);
    return (DataStore)o;
  }

  public void tofile(String fname) throws IOException
  {
    synchronized(this)
    {
      Logger.log("DataStore.java","Writing datastore to disk",Logger.MINOR);
      FileOutputStream stream=new FileOutputStream(fname);
      ObjectOutputStream ostream=new ObjectOutputStream(stream);
      ostream.writeObject(this);
      ostream.close();
    }
  }

  public void callback()
  {
    try {
      tofile(filename);
      Node.timer.add(23, write_interval, this);
    } catch(Exception e) {e.printStackTrace();}
  }

  // Public Fields
  public long maxData;

  // Protected Fields
  protected CyclicArray ar;
  protected Hashtable h;

  // Constructor

  public StandardDataStore(long maxData, int maxItems) {
    ar = new CyclicArray(maxItems);
    h = new Hashtable();
    this.maxData = maxData;
    }

  // Public Methods

  public Enumeration keys()
  {
    return h.keys();
  }

  /**
   * Searches for a key in the DataStore and if found returns
   * any data associated with that key.  If none is found,
   * null is returned.
   * @param k The key to search for
   * @return The data associated with the key or null if not found
   **/

  public Data searchData(Key k) {
    Logger.log("DataStore.java", "searchData("+k + ")", Logger.DEBUG);
    synchronized (this)
    {
      DataStoreItem dsi = (DataStoreItem)h.get(k);
      if(dsi==null || dsi.data==null)
      {
        Logger.log("DataStore.java", "Not found.", Logger.DEBUG);
        return null;
      }
      // We have a hit, raise its priority
      Logger.log("DataStore.java", "Data found.", Logger.DEBUG);
      Data data = dsi.data;
      Address addr = dsi.ref;
      remove(k);
      put(k, addr, data);
      return data;
    }
  }

  /**
    * Searches for a key in the DataStore and if found returns
    * any reference associated with that key.  If none is found,
    * null is returned.
    * @param k The key to search for
    * @return The reference associated with the key or null
    *         if not found
    **/

  public Address searchRef(Key k) {
      if (k == null)
	  return null;
    Logger.log("DataStore.java", "called searchRef("+k + ")", Logger.DEBUG);
    synchronized (this)
    {
      DataStoreItem dsi = (DataStoreItem)h.get(k);
      if(dsi==null)
      {
        Logger.log("DataStore.java", "Not found.", Logger.DEBUG);
        return null;
      }
      Logger.log("DataStore.java", "Ref found.", Logger.DEBUG);
      return dsi.ref;
    }
  }

  /**
      * Removes an item from the DataStore given its key
    * @param k The key of the item to remove
    **/

  public void remove(Key k) {
    synchronized (this) {
      for (int x = 0; x < ar.length(); x++) {
        DataStoreItem dsi = (DataStoreItem) ar.get(x);
        Logger.log("DataStore.java",
                   "Testing "+dsi + " against "+k, Logger.DEBUGGING);
        if (dsi != null)
          if (dsi.key.equals(k)) {
            ar.remove(x);
            h.remove(k);
            return;
            }
        }
      }
    }

  /**
    * Returns the reference associated with the closest key
    * to k.
    * @param k The key to search for
    * @return The closest Key to k
    **/
  public Key findClosestKey(Key k) {
    return findClosestKey(k, null);
    }

  /**
    * Returns the reference associated with the closest key
    * to k.  The masking key allows the 'next best' key to
    * be found.  Imagine the entries in the datastore being
    * sorted in terms of closeness to the key you specify.
    * With no mask key the reference associated with the
    * top-most item will be returned.  However, with a mask
    * key any references associated with keys above the mask
    * key (including the mask key itself) will be ignored.
    * This facility allows best-first backtracking.
    * @param k The key to search for
    * @param maskKey The masking key (or null for no mask)
    * @return The closest Key to k (excluding masked keys) or
    *         null if all keys are masked.
    **/

  public Key findClosestKey(Key k, Key maskKey) {
    int m = 0;
    synchronized (this) {
      // Ensure that maskKey exists
      if ((maskKey != null) && (searchRef(maskKey) == null))
        maskKey = null;
      int indx = findNextClosestPos(k, maskKey);
      if (indx < 0)
	  return null;
      DataStoreItem dsi = (DataStoreItem) ar.get(indx);
      if (dsi == null)
        return null;
      return dsi.key;
      }
    }

  /**
    * Adds an item to the DataStore
    * @param k The key to add
    * @param r The address to which requests should be directed for
    *          this key if the data is removed
    * @param d The data associated with this key
    **/

  public void put(Key k, Address r, Data d) {
    synchronized (this) {
      DataStoreItem dsi = new DataStoreItem(k, r, d);
      DataStoreItem old = (DataStoreItem) ar.put(dsi);
      if (old != null)
	  h.remove(old.key);
      h.put(k, dsi);
      cleanUpData();
      }
    }

  // Protected Methods

  /**
   * Will return the position in ar of the closest key to k that is not
   * as close as nextk
   * @param k Key to search for
   * @param nextk Must not be as close as this key, if non-null
   * @return Position in ar of closest key, -1 if there is no key in ar
   *         such that nextk is closer to k then that key
   **/
  protected int findNextClosestPos(Key k, Key nextk) {
    synchronized (this) {
      int cl = -1;
      DataStoreItem bs = null;
      for (int x = 0; x < ar.length(); x++) {
        DataStoreItem ds = (DataStoreItem) ar.get(x);
        if ((ds != null) &&
            ((bs == null) || (k.compare(ds.key, bs.key)))) {
          // ds is closer than bs
          if ((nextk==null) || (k.compare(nextk, ds.key))) {
            // nextk is closer than ds
            // note that k.compare returns false on a tie
            cl = x;
            bs = ds;
          }
        }
      }
      return cl;
    }
  }

  /**
    * Removes excess data (ie. if there is more data than is
    * permitted by the maxData field).
    **/
  protected void cleanUpData() {
    long total = 0;
    for (int x = 0; x < ar.length(); x++) {
      Object o = ar.get(x);
      if (o != null) {
        DataStoreItem dsi = (DataStoreItem) o;
        if (dsi.data != null) {
          if (total + dsi.data.length() > maxData)
            dsi.data = null;
          else
            total += dsi.data.length();
          }
        }
      }
    }

  // Mainly for debugging
  public String toString()
  {
    return h.toString();
  }
}

class DataStoreItem implements Serializable {
  public Key key;
  public Address ref;
  public Data data;

  public DataStoreItem(Key k, Address a, Data d) {
    key = k;
    ref = a;
    data = d;
    }

  public String toString() {
    return "Key: "+key + " Ref: "+ref + " Data:"+data;
    }
  }

/*
 *$Log: StandardDataStore.java,v $
 *Revision 1.10  2000/04/08 20:48:17  hobbex
 *Write DataStore to .freenet/store_<port>
 *
 *Revision 1.9  2000/04/08 19:19:52  hal
 *synch when dumping datastore to disk so it can be modified in the middle.
 *
 *Revision 1.8  2000/04/08 15:22:19  hobbex
 *Cleaned up my ugly code from yesterday
 *
 *Revision 1.7  2000/04/08 07:12:15  hal
 *Move items to top of priority stack when searchData gets a hit.  Not
 *very efficient.
 *
 *Revision 1.6  2000/04/07 23:19:33  hobbex
 *Remove entry from hash when it is overwritten in the array
 *
 *Revision 1.5  2000/04/07 20:34:09  hobbex
 *Fixed building and running with Kaffe (scripts/kbuild.sh). Added canceling of timer callback and sending of requestfailed on loosing MessageMemory (not too well tested). And fixed a bug in the datastore + some exception catching in requestfailed.
 *
 *Revision 1.4  2000/04/07 06:03:42  hal
 *Clean up StandardDataStore:
 *Make searchRef use the hashtable like searchData, no searching necessary
 *Clean up findClosestKey to only search once not multiple times.
 *Fix findClosestKey, it was not looking at the most recently inserted item.
 *
 *Revision 1.3  2000/04/03 21:04:55  hal
 *Fix findClosestPos() in StandardDataStore.
 *
 *Revision 1.2  2000/04/02 21:13:25  blanu
 *Moved class loading into a separate class (Freenet.support.Loader)
 *
 *Revision 1.1  2000/03/29 18:49:14  hobbex
 *Subclassed Node from Core, added Freenet.node package
 *
 *Revision 1.18  2000/03/25 17:42:40  sanity
 *Minor tidying of debug messages.
 *
 *Revision 1.17  2000/03/22 18:19:36  blanu
 *Added Hashtable for keys for faster lookup.
 *
 *Revision 1.16  2000/03/17 10:06:02  blanu
 *Changed Freenet.support.Timer to Freenet.support.Ticker to avoid JDK1.3 name conflict.
 *
 *Revision 1.15  2000/02/23 20:52:42  blanu
 *Fixed timeout in client.
 *
 *Revision 1.14  2000/02/14 21:08:14  blanu
 *The datastore is now written to disk every checkpointInterval seconds as
 *per the configuration file, with a default of every 5 minutes.
 *
 *Revision 1.13  2000/02/14 00:20:31  blanu
 *Periodic writing of DataStore information to .freenet/store. It doesn't
 *work yet. I think it might be something wrong with Ticker.
 *
 *Revision 1.12  2000/02/13 12:11:57  sanity
 *Fixed numerous syntax errors.
 *
 *Revision 1.11  2000/02/09 23:56:39  hobbex
 *Preliminary tunneling of DataReplies and DataInserts (for now)
 *
 *Revision 1.10  2000/02/08 16:22:06  sanity
 *Removed printing of exception when datastore file is not found, replaced
 *with logger message.
 *
 *Revision 1.9  2000/02/02 21:03:06  blanu
 *Changed the fromfile method so that it should work without a store file.
 *
 *Revision 1.8  2000/02/02 11:26:40  sanity
 *Merged changes.
 *
 *Revision 1.7  2000/02/02 05:13:28  blanu
 *Datastore can now be read to / written from a file. Params can be fetched as various types.
 *
 *Revision 1.6  2000/01/03 10:20:48  hobbex
 *Removed deleting of MM on Loop and Failed Request
 *
 *Revision 1.5  2000/01/03 00:09:03  michaels
 *findClosestPos didn't respect a set ignore flag on first dsi in array (cl==0)
 *now it does.
 *
 *Revision 1.4  2000/01/02 14:10:15  michaels
 *findClosestKey doesnt't return a dsi anymore if its ignore-flag is set to true
 *
 *Revision 1.3  2000/01/02 13:05:06  michaels
 *searchRef is used for checking that a maskKey exists (not searchData)
 *
 */
